#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>

/**
 * 転送データの基本情報
 */

// ターゲットボードの転送アドレス
#define LOAD_ADDR		0xC0008000
// 転送データのヘッダサイズ
#define HEADERSIZE		8
// 転送データのCRCサイズ
#define CRCSIZE			2

/**
 * 転送データ管理情報
 *
 * @file カーネルイメージファイルのファイルディスクリプタ
 * @fileStat カーネルイメージファイルの stat 情報
 * @dev USBデバイスファイルのファイルディスクリプタ
 * @loadAddress 転送先のメモリアドレス
 * @transSize 転送サイズ
 * @CRC 転送データのCRC
 * @rawData 転送データとして加工する前のデータ
 * @transData 転送データ
 */
struct transferDataInfo {
	int				file;
	struct stat     fileStat;
	int				dev;
	unsigned int 	loadAddress;
	unsigned int 	transSize;
	unsigned short 	CRC;
	unsigned char* 	rawData;
	unsigned char* 	transData;
};

/**
 * デストラクタ
 *
 * @param p 解放対象メモリ
 */
static void
deleteTransferDataInfo(struct transferDataInfo *p) {
	if (NULL != p) {
		if (-1 != p->file) {
			close(p->file);
			p->file = -1;
		}
		if (-1 != p->dev) {
			close(p->dev);
			p->dev = -1;
		}
		if (NULL != p->rawData) {
			free(p->rawData);
			p->rawData = NULL;
		}
		if (NULL != p->transData) {
			free(p->transData);
			p->transData = NULL;
		}
		free(p);
	}
}
/**
 * コンストラクタ
 *
 * @param argc コマンド引数の数
 * @param argv コマンド引数
 * @return 成功 転送データ管理情報、失敗 NULL
 */
static struct transferDataInfo *
newTransferDataInfo(int argc, const char* argv[]) {
	struct transferDataInfo *p = NULL;

	// 引数チェック
	if (3 != argc) {
		fprintf(stderr, "Usage: dnw filename devpath\n");
		return NULL;
	}
	p = (struct transferDataInfo*)calloc(1, sizeof(struct transferDataInfo));
	if (!p) {
		fprintf(stderr, "No memory!\n");
		return NULL;
	}
	// 転送対象ファイルのオープン
	p->file = open(argv[1], O_RDONLY);
	if(-1 == p->file) {
		fprintf(stderr, "Can not open file - %s\n", argv[1]);
		goto errdone;
	}

	// USBデバイスファイルのオープン
	p->dev = open(argv[2], O_WRONLY);
	if( -1 == p->dev) {
		fprintf(stderr, "Can not open %s\n", argv[2]);
		goto errdone;
	}
	if ( -1 == fstat(p->file, &p->fileStat)) {
		fprintf(stderr, "Get file size filed!\n");
		goto errdone;
	}
	// ファイル読み込み用のメモリ確保
	p->rawData = (unsigned char*)malloc(p->fileStat.st_size);
	if (NULL == p->rawData) {
		fprintf(stderr, "malloc failed!\n");
		goto errdone;
	}
	// ファイル読み込み
	if (p->fileStat.st_size !=  read(p->file, p->rawData, p->fileStat.st_size)) {
		fprintf(stderr, "Read file failed!\n");
		goto errdone;
	}
	// データ転送用のメモリ確保
	p->transSize = p->fileStat.st_size + HEADERSIZE + CRCSIZE;
	p->transData = (unsigned char*)malloc(p->transSize);
	if (NULL == p->transData) {
		fprintf(stderr, "malloc failed!\n");
		goto errdone;
	}
	p->loadAddress = LOAD_ADDR;
	return p;
errdone:
	deleteTransferDataInfo(p);
	return NULL;
}

/**
 * 転送データのCRC を設定する。
 *
 * @param p 転送データ管理情報
 */
static void
setCRC(struct transferDataInfo *p) {
        int i;
        p->CRC = 0;
        for(i = 0; i < p->fileStat.st_size; i++) {
                p->CRC += p->rawData[i];
        }
}

/**
 * 転送用データを設定する
 *
 * @param p 転送データ管理情報
 */
static void
setTransferData(struct transferDataInfo *p) {
        struct header {
                unsigned int loadAddr;
                unsigned int transSize;
	} *hp = (struct header *)p->transData;
        char *rawData = (char*)&(p->transData[HEADERSIZE]);
        unsigned short *CRC = (unsigned short*)&(p->transData[HEADERSIZE + p->fileStat.st_size]);

        // ヘッダの設定
        hp->loadAddr = p->loadAddress;
        hp->transSize = p->transSize;

        // データ部の設定
        memcpy(rawData, p->rawData, p->fileStat.st_size);

        // CRC の設定
        *CRC = p->CRC;
}

/**
 * USB デバイスにデータ転送する。
 *
 * @param p 転送データ管理情報
 * @return 成功 0、失敗 1
 */
static int
transfer(struct transferDataInfo *p) {
	printf("Writing data...\n");
	size_t remain_size = p->transSize;
	size_t block_size = remain_size / 100;
	size_t writed = 0;
	while(remain_size > 0) {
		size_t to_write = remain_size > block_size ? block_size : remain_size;
		if( to_write != write(p->dev, p->transData + writed, to_write)) {
			fprintf(stderr, "failed!\n");
			return 1;
		}
		remain_size -= to_write;
		writed += to_write;
		fflush(stdout);
	}
	printf("OK\n");
	return 0;
}

/**
 * コマンドの使用方法
 *
 *  dnw カーネルイメージファイル USBデバイスファイル
 */
int
main(int argc, const char* argv[])
{
	int err = 0;
	struct transferDataInfo *infop = NULL;

	infop = newTransferDataInfo(argc, argv);
	if (!infop) {
		return 1;
	}

	setCRC(infop);

	setTransferData(infop);

	err = transfer(infop);
	if (0 != err) {
	    fprintf(stderr, "transfer failed\n");
	}

	deleteTransferDataInfo(infop);

	return err;
}
